searchNodes=[{"ref":"Mine.html","title":"Mine","type":"module","doc":""},{"ref":"Mine.html#append/2","title":"Mine.append/2","type":"macro","doc":"When to_view is invoked, an extra field with the key key will be added to the resulting map with the value value. Useful when an external API is not hygienic and requires constant metadata for an entity to be properly parsed. For example: defmodule MyModule do use Mine struct [:field] defview do append &quot;@class&quot;, &quot;Some.Java.Class&quot; end end MyModule.to_view(%MyModule{field: &quot;example&quot;}) # %{&quot;field&quot; =&gt; &quot;example&quot;, &quot;@class&quot; =&gt; &quot;Some.Java.Class&quot;}"},{"ref":"Mine.html#default_view/1","title":"Mine.default_view/1","type":"macro","doc":"Changes the name of the default view to name. The default view name will be the fallback parameter for any calls to to_view/1 and from_view/1. Normally, these functions will be generated with signatures similar to the following: def to_view(struct = %MyModule{}, view \\\\ :default) def from_view(map, view \\\\ :default) But declaring default_view :other in your module... defmodule MyModule do use Mine # create struct default_view :other defview :other do # ... end end ...yields signatures like these: def to_view(struct = %MyModule{}, view \\\\ :other) def from_view(map, view \\\\ :other)"},{"ref":"Mine.html#defview/2","title":"Mine.defview/2","type":"macro","doc":"Used to specify the requirements for generating to_view and from_view functions on a module. Within the scope of this macro, Mine.field/2, Mine.append/2, and Mine.ignore/1 will be in scope. The name argument is a value used to identify this view when invoking the generated to_view and from_view functions. This defaults to, well, :default. If there are multiple views declared in a module and you wish to change the default behavior of the single arity to_view and from_view functions, use Mine.default_view/1. Note that any fields that exist in the modules struct and are not explicitly ignored with a call to Mine.ignore/1 will not be ignored. This behavior may be configurable in the future. Annotations can be placed before defview to augment the view. @default_view: if passed true, indicates that the following view should be the default. It can also be passed the name of a view, similar to Mine.default_view/1. @exclude_if: should be passed a function with arity 1 that returns true or false. Used to filter values maps resulting from a call to to_view. This is useful for when empty fields should not be included in the final data representation. Two helper functions can be referenced using atoms: :is_nil (entries whose values are nil will be filtered out), and :is_blank (entries whose values are nil or an empty string are removed)."},{"ref":"Mine.html#field/2","title":"Mine.field/2","type":"macro","doc":"Changes the name use to identify a field in an external view. There are many cases where the name you use for a field within your Elixir application is not the name used outside of your application. This is especially apparent when interfacing with an API that defines variables in camel case. Since we conventionally represent atoms (and therefore the keys of structs) in snake case, this can lead to code mapping field names to and from the external naming convention. field/2 is a simpler way to represent this relationship. Options: as: Name to use when mapping to/from an external view default: Value to be used when struct does not contain an entry for the given key map_to: Function with arity 1 that will be invoked on the existing value in the struct when using to_view map_from: Function with arity 1 that will be invoked on the value found in the external view when using from_view Example: defmodule MyModule do use Mine struct [:my_field, :foo, :to_upper] defview do field :my_field, as: &quot;myField&quot; field :foo, default: &quot;bar&quot; field :to_upper, map_to: &amp;String.upcase/1 end end MyModule.to_view(%MyModule{my_field: 1, to_upper: &quot;upper!&quot;}) # %{&quot;myField&quot; =&gt; 1, &quot;foo&quot; =&gt; &quot;bar&quot;, &quot;to_upper&quot; =&gt; &quot;UPPER!&quot;}"},{"ref":"Mine.html#handle_attributes!/2","title":"Mine.handle_attributes!/2","type":"function","doc":""},{"ref":"Mine.html#ignore/1","title":"Mine.ignore/1","type":"macro","doc":"Declares that the given key will not be considered when mapping a struct in either to_view or from_view. For example: defmodule MyModule do use Mine struct [:public, :private] defview do ignore :private end end MyModule.to_view(%MyModule{public: &quot;hello&quot;, private: &quot;world&quot;}) # %{&quot;public&quot; =&gt; &quot;hello&quot;} MyModule.from_view(%{&quot;public&quot; =&gt; &quot;hello&quot;, &quot;private&quot; =&gt; &quot;world&quot;}) # %MyModule{public: &quot;hello&quot;} A field that has been declared as ignored cannot be used in Mine.field/2 and vice versa. Doing so will raise an error."},{"ref":"Mine.html#valid_key?/1","title":"Mine.valid_key?/1","type":"function","doc":""},{"ref":"Mine.html#t:key/0","title":"Mine.key/0","type":"type","doc":""},{"ref":"Mine.View.CompileError.html","title":"Mine.View.CompileError","type":"exception","doc":""},{"ref":"Mine.View.CompileError.html#t:t/0","title":"Mine.View.CompileError.t/0","type":"type","doc":""},{"ref":"readme.html","title":"Mine","type":"extras","doc":"Mine Mine is a lightweight package for defining views for structs. It aims to eliminate the boilerplate required to write and maintain an anti-corruption layer between your structs and any external APIs you interface with. Note: Mine currently exports the minimum functionality required for my own use case but is not ready for adoption yet. Until v1.0.0, the interface may change in ways that break backwards compatibility."},{"ref":"readme.html#usage","title":"Mine - Usage","type":"extras","doc":"Basic Views defmodule User do use Mine # module must declare a struct defstruct [:name, :email, :password] defview do field :name, as: &quot;userName&quot;, default: &quot;&quot; ignore :password end end A view is created by importing Mine at the beginning of your module and using the exported defview/2 macro. Note that User declares a struct and the defview is used after the declaration. This requirement stems from the intent of Mine to be purely additive. It will not replace functionality provided by more mature modules but rather build on existing code. The most notable benefit to this approach is that other packages that internally define structs (like Ecto) can be extended. At compile time, Mine will create views for every module in use. Any incompatibilities between the declared views and their base structs will be caught here, providing the benefit of runtime safety. To access the compiled view at runtime, the following methods will be found on your module. to_view/2: Covert struct to a map for a given view User.to_view(%User{name: &quot;Bob&quot;, email: &quot;abc@d.com&quot;, password: &quot;secret&quot;}, :default) # %{&quot;userName&quot; =&gt; &quot;Bob&quot;, &quot;email&quot; =&gt; &quot;abc@d.com&quot;} from_view/2: Convert source map to struct User.from_view(%{&quot;email&quot; =&gt; &quot;abc@d.com&quot;, &quot;password&quot; =&gt; &quot;shouldn&#39;t be here&quot;}, :default) # %User{name: &quot;&quot;, email: &quot;abc@d.com&quot;, password: nil} The first argument for both of these functions is the data to translate. The second is the name of the view to use for this translation. A module using Mine can have multiple views to capture the formatting requirements of different domains. In this case, :default was not needed, since the second parameter defaults to this value unless configured otherwise. These functions should ideally be used between your serialization layer and business logic. For example, after retrieving data from a third party API, your application would deserialiaze the data using devinus/poison or michalmuskala/jason, pass the resulting map through Mine, and then validate the results using an Ecto changeset before persisting the data. More info on these functions can be found on HexDocs. Available Macros There are several macros that can be used inside the scope of defview. They are: field/2: Change the key for a given field. Has the following uses: field(:key, &quot;as&quot;) View will use &quot;as&quot; instead of the field name :key field(:key, default: &quot;def&quot;) Same as above, but if :key is not found or it&#39;s value is nil, &quot;def&quot; will be used as the value instead. field(:key, map_to: &amp;String.upcase/1) When using to_view, value of field in struct will be accessed, mapped using map_to, and stored under the key &quot;as&quot; in the resulting map field(:key, map_from: &amp;String.downcase/1) When using from_view, value of &quot;as&quot; in the given map will be fetched and passed as the only argument to the function in map_from. ignore/1: Field will be ignored in both to_view and from from_view. append/2: A key value pair that will be added to any map produced by to_view. Ignored in from_view. Named Views In some cases, it may not be enough to have a single exported view of a struct. To support this use case, every view created by Mine has a name. :default will be used when a name is not provided. default_view/1 changes which view is used for the single arity versions of to_view and from_view. A more complicated User module may look something like this: ... # to_veiw/1 and from_view/1 will use this view # instead of :default default_view :front_end # other fields will remain untouched defview :third_party_api do field :name, as: &quot;userName&quot; end defview :front_end do field :name, default: &quot;?&quot; field :email, default: &quot;unknown&quot; ignore :password end # alternatively, use the @default_view annotation @default_view true defview :front_end do # ... end ... After generation, the module will have function definitions similar to the following: def to_view(%User{}, view \\\\ :front_end) def to_view(%User{}, :front_end) def to_view(%User{}, :third_party_api) # plus corresponding from_view functions With this layout, you can let pattern matching determining which view to use."},{"ref":"readme.html#key-naming-strategies","title":"Mine - Key Naming Strategies","type":"extras","doc":"A common requirement for views is to translate struct keys to another namingconvention, as snake case is not always the norm. To support this use case, a view can be annotated with @naming_strategy and by default, the generated aliases will use the new naming convention. defstruct [:field_one, :field_two] @naming_strategy :camel defview do field :field_one, as: &quot;something&quot; # explicit field aliases take precedence end Available naming strategies are: :camel, :constant, :dot, :kebab, :pascal, and :path. Refer to Recase for details."},{"ref":"readme.html#rationale","title":"Mine - Rationale","type":"extras","doc":"While interfacing with an external API written in Java, I frequently ran across instances in which I would need to map a struct&#39;s key, ignore certain fields, or add constant fields to outgoing requests. The process of setting up these mappings is tiresome and potentially error prone. Let&#39;s take a small, real-world example. This is the required JSON format for a port in an unnamed API: { &quot;$&quot;: 7000, &quot;@enabled&quot;: false } Formats such as this were scattered throughout the API, leading to several modules with structures similar to the following: defmodule Port do defstruct [:num, :enabled] def to_view(%Port{num: num, enabled: enabled}) do %{ &quot;$&quot; =&gt; num, &quot;@enabled&quot; =&gt; enabled } end def from_view(%{&quot;$&quot; =&gt; num, &quot;@enabled&quot; =&gt; enabled}) do %Port{num: num, enabled: enabled} end end Code like this: is necessary to maintain a clean internal structure adds more noise than meaning can be error prone Given these conditions, I opted to break my first rule of writing macros in Elixir (avoid them). Instead of the example seen above, using mine allows for a much more concise representation of a mapping to the external world. defmodule Port do use Mine defstruct [:num, :enabled] defview do field :num, &quot;$&quot; field :enabled, &quot;@enabled&quot; end end This representation is easier to read, maintain, and is even checked for validity at compile time. In addition, the generated code is nearly identical to the code written by hand, resulting in a minimal performance impact."},{"ref":"readme.html#installation","title":"Mine - Installation","type":"extras","doc":"Mine can be installed by adding mine to your list of dependencies in mix.exs: def deps do [ {:mine, &quot;~&gt; 0.3.2&quot;} ] end"},{"ref":"readme.html#benchmarking","title":"Mine - Benchmarking","type":"extras","doc":"To benchmark the generated functions against those written by hand, run the following: shell script MIX_ENV=bench mix run bench/run.exs"},{"ref":"changelog.html","title":"Changelog","type":"extras","doc":"Changelog"},{"ref":"changelog.html#v0-3-2","title":"Changelog - v0.3.2","type":"extras","doc":"Improvement Addition of @naming_strategy annotation. Place the annotation above a view definition and the generated keys will be formatted with that strategy. Available naming strategies are: :camel, :constant,:dot, :kebab, :pascal, and :path. Refer to Recase for details."},{"ref":"changelog.html#v0-3-1","title":"Changelog - v0.3.1","type":"extras","doc":"Improvements Addition of two annotations: @default_view and @exclude_if. @exclude_if should be passed a predicate that take a value from a map produced using to_view and returns true if the key/value pair should be excluded. Convenience for removing empty and nil values from a view if needed."},{"ref":"changelog.html#v0-3-0","title":"Changelog - v0.3.0","type":"extras","doc":"Breaking Change Renaming functions to be more concise. alias_field -&gt; field, add_field -&gt; append, ignore_field -&gt; ignore. Larger structs made the verbosity of the previous method names obvious. I apologize for the breaking change, but I think this new language will be more concise. Improvements Can optionally restrict which functions are generated using :only in the use statement. For example, if you only wanted to generate the to_view function, you would define your module like the following: defmodule YourModule do use Mine, only: :to_view # ... rest of module end"},{"ref":"changelog.html#v0-2-3","title":"Changelog - v0.2.3","type":"extras","doc":"Improvements Cleaning up generation logic a bit. Use of Macro.var/2 was potentially unsafe, but all uses have been addressed. Explicitly limiting Mine macros allowed within defview. from_view/2 will now only match on maps. Previous implementation matched everything, which 1) caused errors inside Mine if a map was not passed, 2) blocked client modules from defining additional from_view/2 methods. This appears to have a small performance impact, but the safety is worth it. Other Getting reliable benchmarks is still an issue. The generated code is identical to that written by hand, yet the performance seems to vary up to 60% in both directions. Sometimes vanilla Elixir is faster, sometimes they are the same, sometimes, Mine is faster. Disabling power management, closing all other programs, and manually triggering garbage collection between scenarios did not improve reliability."},{"ref":"changelog.html#v0-2-2","title":"Changelog - v0.2.2","type":"extras","doc":"Improvements Fixing scope issues. Previously, map_to and map_from were not able to access imported or required functions. Now all functions that are valid in the scope of a normal function will be available for use within defview. Adding documentation for important macros. Top level documentation for Mine module is still lacking. Switched to using Macro.struct! to determine whether the host module has declared a struct. These failures will now be CompileError instead of Mine.CompileError. Lots of cleanup on the underlying implementation. Removed the use of processes to keep track of current progress in favor of writing to module attributes. Should increase stability and speed up compilation a bit."},{"ref":"changelog.html#v0-2-1","title":"Changelog - v0.2.1","type":"extras","doc":"New Feature: Mapping Functions map_to and map_from. When an anonymous (or captured) function with arity/1 is supplied to alias_field using the mentioned key, values will be mapped after translation. # Example module using mapping feature defmodule A do use Mine defstruct [:b] defview do alias_field :b, as: &quot;B&quot;, map_to: &amp;String.upcase/1 end end Functions given to map_to and map_from maintain the scope of the caller. Private functions and aliases are honored. Imported functions are currently not supported. Values supplied as mappers are trivially validated to be functions. Improvements Macro hygiene. alias_field, ignore_field, etc are now scoped to defview. If as option is not passed to alias_field, the key name will be turned into a string and used instead. Fixed edge case where aliasing after a defview would complain because the view_agent was still registered but not alive. # example defmodule C do use Mine # defines struct defview do # body end # would randomly cause a crash ignore_field(:some_field) end"},{"ref":"from_view_0-2-3_report.html","title":"Benchmark - from_view","type":"extras","doc":"Benchmark - from_view Benchmark run from 2020-03-10 16:06:32.622146Z UTC"},{"ref":"from_view_0-2-3_report.html#system","title":"Benchmark - from_view - System","type":"extras","doc":"Benchmark suite executing on the following system: Operating System Linux CPU Information Intel(R) Core(TM) i7-6500U CPU @ 2.50GHz Number of Available Cores 4 Available Memory 7.70 GB Elixir Version 1.10.1 Erlang Version 22.2.7"},{"ref":"from_view_0-2-3_report.html#configuration","title":"Benchmark - from_view - Configuration","type":"extras","doc":"Benchmark suite executing with the following configuration: :time 5 s :parallel 1 :warmup 2 s"},{"ref":"from_view_0-2-3_report.html#statistics","title":"Benchmark - from_view - Statistics","type":"extras","doc":"Input: empty map Run Time Name IPS Average Devitation Median 99th&nbsp;% Elixir - Functions with Arity 1 11.19 M 89.37 ns ±51108.28% 0 ns 168 ns Elixir - Functions with Arity >1 10.76 M 92.93 ns ±55900.85% 0 ns 175 ns Mine 10.58 M 94.51 ns ±52634.76% 0 ns 114 ns Comparison Name IPS Slower Elixir - Functions with Arity 1 11.19 M &nbsp; Elixir - Functions with Arity >1 10.76 M 1.04x Mine 10.58 M 1.06x Memory Usage Name Memory Factor Elixir - Functions with Arity 1 48 B &nbsp; Elixir - Functions with Arity >1 48 B 1.0x Mine 48 B 1.0x Input: expected Run Time Name IPS Average Devitation Median 99th&nbsp;% Elixir - Functions with Arity >1 11.48 M 87.14 ns ±47526.28% 0 ns 95 ns Elixir - Functions with Arity 1 10.38 M 96.38 ns ±55692.71% 0 ns 296 ns Mine 9.23 M 108.37 ns ±41457.11% 19 ns 219 ns Comparison Name IPS Slower Elixir - Functions with Arity >1 11.48 M &nbsp; Elixir - Functions with Arity 1 10.38 M 1.11x Mine 9.23 M 1.24x Memory Usage Name Memory Factor Elixir - Functions with Arity >1 48 B &nbsp; Elixir - Functions with Arity 1 48 B 1.0x Mine 48 B 1.0x"},{"ref":"to_view_0-2-3_report.html","title":"Benchmark - to_view","type":"extras","doc":"Benchmark - to_view Benchmark run from 2020-03-10 16:05:15.048591Z UTC"},{"ref":"to_view_0-2-3_report.html#system","title":"Benchmark - to_view - System","type":"extras","doc":"Benchmark suite executing on the following system: Operating System Linux CPU Information Intel(R) Core(TM) i7-6500U CPU @ 2.50GHz Number of Available Cores 4 Available Memory 7.70 GB Elixir Version 1.10.1 Erlang Version 22.2.7"},{"ref":"to_view_0-2-3_report.html#configuration","title":"Benchmark - to_view - Configuration","type":"extras","doc":"Benchmark suite executing with the following configuration: :time 5 s :parallel 1 :warmup 2 s"},{"ref":"to_view_0-2-3_report.html#statistics","title":"Benchmark - to_view - Statistics","type":"extras","doc":"Input: empty Run Time Name IPS Average Devitation Median 99th&nbsp;% Elixir - Functions with Arity 1 18.46 M 54.16 ns ±18333.91% 17 ns 170 ns Mine 12.43 M 80.47 ns ±19567.90% 35 ns 317 ns Elixir - Functions with Arity >1 10.56 M 94.69 ns ±17261.83% 58 ns 398 ns Comparison Name IPS Slower Elixir - Functions with Arity 1 18.46 M &nbsp; Mine 12.43 M 1.49x Elixir - Functions with Arity >1 10.56 M 1.75x Memory Usage Name Memory Factor Elixir - Functions with Arity 1 40 B &nbsp; Mine 40 B 1.0x Elixir - Functions with Arity >1 40 B 1.0x Input: expected Run Time Name IPS Average Devitation Median 99th&nbsp;% Mine 3.94 M 253.98 ns ±8388.92% 213 ns 403 ns Elixir - Functions with Arity 1 2.90 M 344.94 ns ±7796.64% 272 ns 981 ns Elixir - Functions with Arity >1 2.68 M 373.64 ns ±7120.25% 296 ns 928 ns Comparison Name IPS Slower Mine 3.94 M &nbsp; Elixir - Functions with Arity 1 2.90 M 1.36x Elixir - Functions with Arity >1 2.68 M 1.47x Memory Usage Name Memory Factor Mine 136 B &nbsp; Elixir - Functions with Arity 1 136 B 1.0x Elixir - Functions with Arity >1 136 B 1.0x"}]